#! /usr/bin/env python3

from collections import OrderedDict
from colorama import Fore, Style
import argparse
import colorama
import functools
import json
import os
import os.path
import random
import re
import string
import subprocess
import sys
import tempfile
import time

DRY_RUN = False

def error(message):
    print(f'{Fore.RED}{Style.BRIGHT}error:{Style.RESET_ALL} {message}', file=sys.stderr)
    sys.exit(1)

def print_cmd(cmd):
    pretty = ' '.join(['"'+c+'"' if ' ' in c else c for c in map(str, cmd)])
    print(f'{Fore.YELLOW}{pretty}{Style.RESET_ALL}')

def run_cmd(cmd, input=None, env=None, retry_fn=None):
    cmd = [str(c) for c in cmd]

    if input:
        input = input.encode()
    if env:
        e = os.environ.copy()
        e.update(env)
        env = e
    stdout = None
    if retry_fn:
        stdout = subprocess.PIPE

    max_tries = 6

    while True:
        try:
            print_cmd(cmd)
            if DRY_RUN:
                return
            proc = subprocess.run(cmd, input=input, stdout=stdout, env=env, check=True)
            if stdout is not None:
                output = proc.stdout.decode()
                print(output, end='')
        except subprocess.CalledProcessError as e:
            output = ''
            if e.output:
                output = e.output.decode()
            if retry_fn is not None and retry_fn(output) and max_tries > 0:
                print('Retrying...')
                max_tries -= 1
                time.sleep(1)
                continue
            error('command failed')
        return

def random_worktree():
    while True:
        path = '/tmp/prpy_' + ''.join(random.choice(string.ascii_lowercase + string.digits)
            for _ in range(8))
        if not os.path.exists(path):
            return path

def enter_worktree():
    old_path = os.path.abspath(os.getcwd())
    new_path = random_worktree()

    run_cmd([
        'git', 'worktree', 'add', '--no-checkout', new_path
        ])

    print_cmd(['cd', new_path])
    if not DRY_RUN:
        os.chdir(new_path)

    return old_path

def leave_worktree(old_path):
    new_path = os.path.abspath(os.getcwd())

    print_cmd(['cd', old_path])
    os.chdir(old_path)

    run_cmd([
        'git', 'worktree', 'remove', '-f', os.path.basename(new_path)
        ])

def remember_ref():
    if DRY_RUN:
        return 'none'
    try:
        output = subprocess.run(
            ['git', 'symbolic-ref', '--short', 'HEAD'],
            stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, check=True).stdout
    except:
        output = subprocess.run(
            ['git', 'rev-parse', 'HEAD'],
            stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, check=True).stdout
    return output.decode().strip()

def restore_ref(ref):
    if os.path.exists('.git/index.lock'):
        run_cmd(['git', 'rebase', '--abort'])

    run_cmd(['git', 'checkout', ref])

def delete_ref(ref):
    run_cmd(['git', 'branch', '-D', ref])

def guess_issue(org, repo, text):
    if not text:
        return None

    delim = r'[,:;?!()\[\]|+*_~<> \t\n\r]'

    prefix = f'(?:^|(?<={delim}))'
    suffix = f'(?:$|(?={delim}))'

    patterns = [
        r'(?:#|gh-)(\d+)',
        r'([\w-]+)/([\w-]+)#(\d+)',
        r'https?://github.com/([\w-]+)/([\w-]+)/issues/(\d+)(?:#[\w\d-]+)?',
    ]

    regexp = re.compile('|'.join([prefix + p + suffix for p in patterns]),
        re.IGNORECASE | re.M)

    for m in regexp.finditer(text):
        if m.group(1):
            return org, repo, int(m.group(1))

        if m.group(2) == org:
            return m.group(2), m.group(3), int(m.group(4))

        if m.group(5) == org:
            return m.group(5), m.group(6), int(m.group(7))

    return None

def make_prefix(org, repo, issue_link):
    if not issue_link:
        error("can't determine issue associated with pr\n"
              "add issue number to pr description or use --issue or --no-issue")

    issue_org, issue_repo, issue_number = issue_link

    if issue_org == org and issue_repo == repo:
        return f'gh-{issue_number}'
    else:
        return f'{issue_org}/{issue_repo}#{issue_number}'

def make_message(org, repo, issue_link, pr_title):
    pr_title = re.sub(r'\s*\(#\d+\)$', '', pr_title)
    pr_title = re.sub(r'\.$', '', pr_title)

    if issue_link:
        return '{} {}'.format(
            make_prefix(org, repo, issue_link),
            pr_title)
    else:
        return pr_title

@functools.cache
def query_issue_info(org, repo, issue_number):
    issue_info = {}

    try:
        response = json.loads(subprocess.run([
            'gh', 'api',
            f'/repos/{org}/{repo}/issues/{issue_number}'
            ],
            capture_output=True, text=True, check=True).stdout)
    except subprocess.CalledProcessError as e:
        error(f'failed to retrieve issue info: {e.stderr.strip()}')

    issue_info['issue_title'] = response['title']
    issue_info['issue_url'] = response['html_url']

    if response['milestone']:
        issue_info['issue_milestone'] = response['milestone']['title']
    else:
        issue_info['issue_milestone'] = None

    return issue_info

@functools.cache
def query_pr_info(org, repo, pr_number, no_git=False):
    try:
        response = json.loads(subprocess.run(
            ['gh', 'api', f'/repos/{org}/{repo}/pulls/{pr_number}'],
            capture_output=True, text=True, check=True).stdout)
    except subprocess.CalledProcessError as e:
        error(f'failed to retrieve pr info: {e.stderr.strip()}')

    pr_info = {
        'pr_link': (org, repo, pr_number),
        'pr_title': response['title'],
        'pr_url': response['html_url'],
        'pr_state': response['state'],
        'pr_draft': response['draft'],
        'pr_mergeable': response['mergeable'],
        'pr_rebaseable': response['rebaseable'],
        'source_branch': response['head']['ref'],
        'source_sha': response['head']['sha'],
        'source_remote': response['head']['repo']['ssh_url'],
        'target_branch': response['base']['ref'],
        'target_sha': response['base']['sha'],
        'target_remote': response['base']['repo']['ssh_url'],
    }

    if response['milestone']:
        pr_info['pr_milestone'] = response['milestone']['title']
    else:
        pr_info['pr_milestone'] = None

    pr_info['issue_link'] = None

    if 'body' in response:
        pr_info['issue_link'] = pr_info['issue_link_in_body'] = \
          guess_issue(org, repo, response['body'])

    if not pr_info['issue_link'] and 'title' in response:
        pr_info['issue_link'] = guess_issue(org, repo, response['title'])

    if not pr_info['issue_link'] and 'issue' in response:
        pr_info['issue_link'] = (org, repo, int(response['issue']['number']))

    if pr_info['issue_link']:
        issue_info = query_issue_info(*pr_info['issue_link'])
        pr_info.update(issue_info)

    if not no_git:
        try:
            pr_info['base_sha'], pr_info['base_ref'] = subprocess.run(
                ['git', 'ls-remote', pr_info['target_remote'], pr_info['target_branch']],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True).stdout.decode().strip().split()
        except subprocess.CalledProcessError as e:
            error(f'failed to retrieve git remote info: {e.stderr.decode().strip()}')

    return pr_info

@functools.cache
def query_pr_actions(org, repo, pr_number, no_git=False):
    pr_info = query_pr_info(org, repo, pr_number, no_git)

    try:
        response = json.loads(subprocess.run([
            'gh', 'api',
            f'/repos/{org}/{repo}/actions/runs?event=pull_request',
            ],
            capture_output=True, text=True, check=True).stdout)
    except subprocess.CalledProcessError as e:
        error(f'failed to retrieve workflow runs: {e.stderr.strip()}')

    results = {}
    for run in response['workflow_runs']:
        if run['head_sha'] != pr_info['source_sha']:
            continue
        status = run['status']
        if status == 'completed':
            status = run['conclusion']
        results[run['name']] = status

    return sorted(results.items())

@functools.cache
def query_pr_commits(org, repo, pr_number, no_git=False):
    try:
        response = json.loads(subprocess.run([
            'gh', 'pr', 'view',
            '--repo', f'{org}/{repo}',
            '--json', 'commits',
            str(pr_number),
            ],
            capture_output=True, text=True, check=True).stdout)
    except subprocess.CalledProcessError as e:
        error(f'failed to retrieve pr commits: {e.stderr.strip()}')

    results = []
    for commit in response['commits']:
        results.append((commit['oid'],
                commit['messageHeadline'],
                commit['authors'][0]['name'], commit['authors'][0]['email']))

    return results

def show_pr(org, repo, pr_number, show_json):
    pr_info = query_pr_info(org, repo, pr_number, no_git=True)
    pr_actions = query_pr_actions(org, repo, pr_number, no_git=True)
    pr_commits = query_pr_commits(org, repo, pr_number, no_git=True)

    json_result = OrderedDict()
    json_section = {}

    def section(name, ctor=OrderedDict):
        if show_json:
            nonlocal json_result, json_section
            json_section = ctor()
            json_result[name.replace(' ', '_')] = json_section
        else:
            print(f'{Fore.GREEN}{Style.BRIGHT}{name}:{Style.RESET_ALL}')

    def keyval(key, val, color=None):
        if show_json:
            nonlocal json_section
            json_section[key] = val
        else:
            if color:
                print(f'    {key}: {color}{Style.BRIGHT}{val}{Style.RESET_ALL}')
            else:
                print(f'    {key}: {val}')

    def empty(val, color=None):
        if show_json:
            pass
        else:
            if color:
                print(f'  {color}{Style.BRIGHT}{val}{Style.RESET_ALL}')
            else:
                print(f'  {val}')

    def commit(sha, msg, author, email):
        if show_json:
            nonlocal json_section
            json_section.append({
                'sha': sha,
                'message': msg,
                'author': author,
                'email': email,
            })
        else:
            if 'users.noreply.github.com' in email:
                email = 'noreply.github.com'
            print(f'  {sha[:8]} {Fore.BLUE}{Style.BRIGHT}{msg}{Style.RESET_ALL}'+
                f' ({author} <{email}>)')

    def end():
        if show_json:
            print(json.dumps(json_result, indent=2))

    section('pull request')
    keyval('title', pr_info['pr_title'], Fore.BLUE)
    keyval('url', pr_info['pr_url'])
    keyval('milestone', str(pr_info['pr_milestone']).lower(),
           Fore.MAGENTA if pr_info['pr_milestone'] is not None else Fore.RED)
    keyval('source', pr_info['source_branch'], Fore.CYAN)
    keyval('target', pr_info['target_branch'], Fore.CYAN)
    keyval('state', pr_info['pr_state'],
          Fore.MAGENTA if pr_info['pr_state'] == 'open' else Fore.RED)
    keyval('draft', str(pr_info['pr_draft']).lower(),
          Fore.MAGENTA if not pr_info['pr_draft'] else Fore.RED)
    keyval('mergeable', str(pr_info['pr_mergeable'] \
                            if pr_info['pr_mergeable'] is not None else 'unknown').lower(),
          Fore.MAGENTA if pr_info['pr_mergeable'] == True else Fore.RED)
    keyval('rebaseable', str(pr_info['pr_rebaseable'] \
                            if pr_info['pr_rebaseable'] is not None else 'unknown').lower(),
          Fore.MAGENTA if pr_info['pr_rebaseable'] == True else Fore.RED)

    section('issue')
    if pr_info['issue_link']:
        keyval('title', pr_info['issue_title'], Fore.BLUE)
        keyval('url', pr_info['issue_url'])
        keyval('milestone', str(pr_info['issue_milestone']).lower(),
               Fore.MAGENTA if pr_info['issue_milestone'] is not None else Fore.RED)
    else:
        empty('not found', Fore.RED)

    section('actions')
    has_actions = False
    for action_name, action_result in pr_actions:
        has_actions = True
        keyval(action_name, action_result,
              Fore.MAGENTA if action_result == 'success' else Fore.RED)
    if not has_actions:
        empty('not found', Fore.RED)

    section('commits', list)
    has_commits = False
    for commit_sha, commit_msg, commit_author, commit_email in pr_commits:
        has_commits = True
        commit(commit_sha, commit_msg, commit_author, commit_email)
    if not has_commits:
        empty('not found', Fore.RED)

    end()

def verify_pr(org, repo, pr_number, issue_number, issue_miletsone, no_checks,
              no_issue, no_milestone):
    pr_info = query_pr_info(org, repo, pr_number)

    if not no_issue:
        if issue_number:
            issue_info = query_issue_info(org, repo, issue_number)
        else:
            issue_info = pr_info

        if not issue_number and not pr_info['issue_link']:
            error("can't determine issue associated with pr\n"
                  "add issue number to pr description or use --issue or --no-issue")

        if not no_milestone:
            if not issue_miletsone and not issue_info['issue_milestone']:
                error("can't determine milestone associated with issue\n"
                      "assign milestone to issue or use --milestone or --no-milestone")

    if pr_info['pr_state'] != 'open':
        error("can't proceed on non-open pr")

    if pr_info['pr_draft']:
        error("can't proceed on draft pr")

    if not no_checks:
        for action_name, action_result in query_pr_actions(org, repo, pr_number):
            if action_result != 'success':
                error("can't proceed on pr with failed checks\n"
                      "use --no-checks to proceed anyway")

def checkout_pr(org, repo, pr_number):
    pr_info = query_pr_info(org, repo, pr_number)

    local_branch = os.path.basename(os.getcwd())
    target_branch = pr_info['target_branch']

    run_cmd([
        'gh', 'pr', 'checkout',
        '--repo', f'{org}/{repo}',
        '-f',
        '-b', local_branch,
        pr_number,
        ])

def update_pr(org, repo, pr_number, issue_number, issue_milestone,
              no_issue, no_milestone):
    def update_link_to_issue():
        pr_info = query_pr_info(org, repo, pr_number)

        nonlocal issue_number
        if not issue_number:
            if pr_info['issue_link']:
                _, _, issue_number = pr_info['issue_link']

        is_uptodate = pr_info['issue_link'] and \
            (pr_info['issue_link'] == pr_info['issue_link_in_body']) and \
            (pr_info['issue_link'] == (org, repo, issue_number))

        if is_uptodate:
            return

        try:
            response = json.loads(subprocess.run(
                ['gh', 'api', f'/repos/{org}/{repo}/pulls/{pr_number}'],
                capture_output=True, text=True, check=True).stdout)
        except subprocess.CalledProcessError as e:
            error(f'failed to retrieve pr info: {e.stderr.strip()}')

        body = '{}\n\n{}'.format(
            make_prefix(org, repo, (org, repo, issue_number)),
            (response['body'] or '').strip()).strip()

        run_cmd([
            'gh', 'pr', 'edit',
            '--repo', f'{org}/{repo}',
            '--body-file', '-',
            pr_number,
            ],
            input=body)

        query_pr_info.cache_clear()

    def update_milestone_of_issue():
        pr_info = query_pr_info(org, repo, pr_number)

        is_uptodate = pr_info['issue_milestone'] and \
            (not issue_milestone or pr_info['issue_milestone'] == issue_milestone)

        if is_uptodate:
            return

        issue_org, issue_repo, issue_number = pr_info['issue_link']

        run_cmd([
            'gh', 'issue', 'edit',
            '--repo', f'{issue_org}/{issue_repo}',
            '--milestone', issue_milestone,
            issue_number,
            ])

        query_issue_info.cache_clear()
        query_pr_info.cache_clear()

    def update_milestone_of_pr():
        pr_info = query_pr_info(org, repo, pr_number)

        is_uptodate = pr_info['pr_milestone'] and \
            pr_info['pr_milestone'] == pr_info['issue_milestone']

        if is_uptodate:
            return

        run_cmd([
            'gh', 'pr', 'edit',
            '--repo', f'{org}/{repo}',
            '--milestone', pr_info['issue_milestone'],
            pr_number,
            ])

        query_pr_info.cache_clear()

    if not no_issue:
        update_link_to_issue()

        if not no_milestone:
            update_milestone_of_issue()
            update_milestone_of_pr()

def fetch_pr(org, repo, pr_number):
    pr_info = query_pr_info(org, repo, pr_number)

    run_cmd([
        'git', 'fetch', pr_info['target_remote'], pr_info['base_sha'],
        ])

def rebase_pr(org, repo, pr_number):
    pr_info = query_pr_info(org, repo, pr_number)

    run_cmd([
        'git', 'rebase', '--onto', pr_info['base_sha'], pr_info['target_sha'],
        ])

def link_pr(org, repo, pr_number, action, no_issue):
    pr_info = query_pr_info(org, repo, pr_number)

    if no_issue:
        commit_prefix = ''
    elif action == 'link':
        commit_prefix = make_prefix(org, repo, pr_info['issue_link']) + ' '
    elif action == 'unlink':
        commit_prefix = ''

    base_sha = pr_info['base_sha']

    run_cmd([
        'git', 'filter-branch', '-f', '--msg-filter',
        f"sed -r -e '1s,^(gh-[0-9]+ +|{org}/[^ ]+ +|[Ii]ssue *[0-9]+:? +)?,{commit_prefix},'"+
          " -e '1s,\s*\(#[0-9]+\)$,,'",
        f'{base_sha}..HEAD',
        ],
        env={'FILTER_BRANCH_SQUELCH_WARNING':'1'})

def squash_pr(org, repo, pr_number, no_issue):
    pr_info = query_pr_info(org, repo, pr_number)

    commit_message = make_message(
        org, repo,
        pr_info['issue_link'] if not no_issue else None,
        pr_info['pr_title'])

    run_cmd([
        'git', 'rebase', '-i', pr_info['base_sha'],
        ],
        env={
            'GIT_EDITOR': ':',
            'GIT_SEQUENCE_EDITOR': "sed -i '1 ! s,^pick,fixup,g'",
        })

    run_cmd([
        'git', 'commit', '--amend', '--no-edit', '-m', commit_message,
        ])

def log_pr(org, repo, pr_number):
    pr_info = query_pr_info(org, repo, pr_number)

    base_sha = pr_info['base_sha']

    run_cmd([
        'git', 'log', '--format=%h %s (%an <%ae>)',
        f'{base_sha}..HEAD',
        ])

def push_pr(org, repo, pr_number):
    pr_info = query_pr_info(org, repo, pr_number)

    local_branch = os.path.basename(os.getcwd())
    source_branch = pr_info['source_branch']
    source_remote = pr_info['source_remote']

    run_cmd([
        'git', 'push', '-f',
        source_remote,
        f'{local_branch}:{source_branch}'
        ])

def wait_pr(org, repo, pr_number):
    while True:
        query_pr_info.cache_clear()

        pr_info = query_pr_info(org, repo, pr_number)
        if pr_info['pr_mergeable'] is not None \
            and pr_info['pr_rebaseable'] is not None:
            break

        time.sleep(0.1)

def merge_pr(org, repo, pr_number):
    def retry_fn(output):
        return 'GraphQL: Base branch was modified' in output or \
            'GraphQL: Pull Request is not mergeable' in output

    time.sleep(1)

    run_cmd([
        'gh', 'pr', 'merge',
        '--repo', f'{org}/{repo}',
        '--rebase',
        '--delete-branch',
        pr_number,
        ],
        retry_fn=retry_fn)

parser = argparse.ArgumentParser(prog='pr.py')

common_parser = argparse.ArgumentParser(add_help=False)
common_parser.add_argument('--org', default='roc-streaming',
                           help='github org')
common_parser.add_argument('--repo', default='roc-toolkit',
                           help='github repo')

action_parser = argparse.ArgumentParser(add_help=False)
action_parser.add_argument('--issue', type=int, dest='issue_number',
                    help="overwrite issue to link with")
action_parser.add_argument('--no-issue', action='store_true', dest='no_issue',
                    help="don't link issue")
action_parser.add_argument('-m', '--milestone', type=str, dest='milestone_name',
                    help="overwrite issue milestone")
action_parser.add_argument('-M', '--no-milestone', action='store_true', dest='no_milestone',
                    help="don't set issue milestone")
action_parser.add_argument('--no-checks', action='store_true', dest='no_checks',
                    help="proceed even if pr checks are failed")
action_parser.add_argument('--no-push', action='store_true', dest='no_push',
                    help="don't actually push anything")
action_parser.add_argument('-n', '--dry-run', action='store_true', dest='dry_run',
                    help="don't actually run commands, just print them")

subparsers = parser.add_subparsers(dest='command')

show_parser = subparsers.add_parser(
    'show', parents=[common_parser],
    help="show pull request info")
show_parser.add_argument('pr_number', type=int)
show_parser.add_argument('--json', action='store_true', dest='json',
                         help="output in json format")

rebase_parser = subparsers.add_parser(
    'rebase', parents=[common_parser, action_parser],
    help="rebase pull request on base branch (keeps it open)")
rebase_parser.add_argument('pr_number', type=int)

link_parser = subparsers.add_parser(
    'link', parents=[common_parser, action_parser],
    help="link pull request description and commits to issue")
link_parser.add_argument('pr_number', type=int)

unlink_parser = subparsers.add_parser(
    'unlink', parents=[common_parser, action_parser],
    help="unlink pull request commits from issue")
unlink_parser.add_argument('pr_number', type=int)

merge_parser = subparsers.add_parser(
    'merge', parents=[common_parser, action_parser],
    help="link and merge pull request")
merge_parser.add_argument('--rebase', action='store_true',
                          help='merge using rebase')
merge_parser.add_argument('--squash', action='store_true',
                          help='merge using squash')
merge_parser.add_argument('pr_number', type=int)

args = parser.parse_args()

if hasattr(args, 'dry_run'):
    DRY_RUN = args.dry_run

colorama.init()

if args.command == 'show':
    show_pr(args.org, args.repo, args.pr_number, args.json)
    exit(0)

if args.command == 'rebase':
    orig_path = enter_worktree()
    pushed = False
    try:
        checkout_pr(args.org, args.repo, args.pr_number)
        pr_ref = remember_ref()
        fetch_pr(args.org, args.repo, args.pr_number)
        rebase_pr(args.org, args.repo, args.pr_number)
        log_pr(args.org, args.repo, args.pr_number)
        if not args.no_push:
            push_pr(args.org, args.repo, args.pr_number)
            pushed = True
    finally:
        leave_worktree(orig_path)
        if pushed:
            delete_ref(pr_ref)
    exit(0)

if args.command == 'link' or args.command == 'unlink':
    verify_pr(args.org, args.repo, args.pr_number, args.issue_number,
              args.milestone_name, args.no_checks, args.no_issue, args.no_milestone)
    orig_path = enter_worktree()
    pushed = False
    try:
        checkout_pr(args.org, args.repo, args.pr_number)
        pr_ref = remember_ref()
        update_pr(args.org, args.repo, args.pr_number, args.issue_number,
                  args.milestone_name, args.no_issue, args.no_milestone)
        fetch_pr(args.org, args.repo, args.pr_number)
        rebase_pr(args.org, args.repo, args.pr_number)
        link_pr(args.org, args.repo, args.pr_number, args.command, args.no_issue)
        log_pr(args.org, args.repo, args.pr_number)
        if not args.no_push:
            push_pr(args.org, args.repo, args.pr_number)
            pushed = True
    finally:
        leave_worktree(orig_path)
        if pushed:
            delete_ref(pr_ref)
    exit(0)

if args.command == 'merge':
    if int(bool(args.rebase)) + int(bool(args.squash)) != 1:
        error("either --rebase or --squash should be specified")
    verify_pr(args.org, args.repo, args.pr_number, args.issue_number,
              args.milestone_name, args.no_checks, args.no_issue, args.no_milestone)
    orig_path = enter_worktree()
    merged = False
    try:
        checkout_pr(args.org, args.repo, args.pr_number)
        pr_ref = remember_ref()
        update_pr(args.org, args.repo, args.pr_number, args.issue_number,
                  args.milestone_name, args.no_issue, args.no_milestone)
        fetch_pr(args.org, args.repo, args.pr_number)
        rebase_pr(args.org, args.repo, args.pr_number)
        if args.rebase:
            link_pr(args.org, args.repo, args.pr_number, 'link', args.no_issue)
        else:
            squash_pr(args.org, args.repo, args.pr_number, args.no_issue)
        log_pr(args.org, args.repo, args.pr_number)
        if not args.no_push:
            push_pr(args.org, args.repo, args.pr_number)
            wait_pr(args.org, args.repo, args.pr_number)
            merge_pr(args.org, args.repo, args.pr_number)
            merged = True
    finally:
        leave_worktree(orig_path)
        if merged:
            delete_ref(pr_ref)
    exit(0)



import SCons.Script
import os
import shutil

def GetDistPath(env, instdir, *args):
    if env.HasArgument('DESTDIR'):
        seps = os.sep+os.altsep if os.altsep else os.sep
        instdir = os.path.join(
            env['DESTDIR'],
            os.path.splitdrive(instdir)[1].lstrip(seps))

    return os.path.join(*([instdir] + list(args)))

def AddDistFile(env, instdir, target):
    if isinstance(target, list):
        target = target[0]

    if isinstance(target, str):
        try:
            target = env.Dir(target)
        except:
            target = env.File(target)

    src = target.path
    dst = env.GetDistPath(instdir, target.name)

    def _install(target, source, env):
        if os.path.isdir(src):
            if os.path.isdir(dst):
                shutil.rmtree(dst)
            shutil.copytree(src, dst)
        elif os.path.islink(src):
            if os.path.isfile(dst):
                os.remove(dst)
            os.symlink(os.readlink(src), dst)
        else:
            par = os.path.dirname(dst)
            if not os.path.exists(par):
                os.makedirs(par)
            shutil.copy(src, dst)

    def _uninstall(target, source, env):
        if os.path.exists(dst):
            if os.path.isdir(dst):
                shutil.rmtree(dst)
            else:
                os.remove(dst)

    env.AlwaysBuild(env.Alias('install', [], [
        env.Action(_install,
                   env.PrettyCommand('INSTALL', dst, 'yellow', 'install({})'.format(dst)))
    ]))

    env.AlwaysBuild(env.Alias('uninstall', [], [
        env.Action(_uninstall,
                   env.PrettyCommand('UNINSTALL', dst, 'red', 'uninstall({})'.format(dst)))
    ]))

def AddDistAction(env, action):
    env.AlwaysBuild(env.Alias('install', [], action))

def init(env):
    env.AlwaysBuild(env.Alias('install', [], env.Action('')))
    env.AlwaysBuild(env.Alias('uninstall', [], env.Action('')))
    env.AddMethod(GetDistPath, 'GetDistPath')
    env.AddMethod(AddDistFile, 'AddDistFile')
    env.AddMethod(AddDistAction, 'AddDistAction')



from __future__ import print_function

import argparse
import errno
import os
import os.path
import re
import subprocess
import sys

parser = argparse.ArgumentParser(description='filter sphinx and doxygen console output')

parser.add_argument('--root-dir', dest='root_dir', type=str, required=True,
                    help='project directory')

parser.add_argument('--work-dir', dest='work_dir', type=str, required=True,
                    help='working directory')

parser.add_argument('--out-dirs', dest='out_dirs', type=str, nargs='+', required=True,
                    help='output directories')

parser.add_argument('--touch-file', dest='touch_file', type=str, required=True,
                    help='file to touch after everything is done')

parser.add_argument('--werror', dest='werror', action='store_true',
                    help='convert warnings to errors')

parser.add_argument('command', nargs=argparse.REMAINDER,
                    help='command to be executed')

if '--' not in sys.argv or sys.argv.index('--') == len(sys.argv)-1:
    parser.print_usage()
    print("error: command is required", file=sys.stderr)
    exit(1)

options = sys.argv[1:sys.argv.index('--')]
command = sys.argv[sys.argv.index('--')+1:]

args = parser.parse_args(options)

project_dir = os.path.abspath(args.root_dir)
working_dir = os.path.abspath(args.work_dir)

for out_dir in args.out_dirs:
    try:
        os.makedirs(out_dir)
    except OSError:
        e = sys.exc_info()[1]
        if e.errno != errno.EEXIST:
            print("error: unable to create '{}' directory".format(out_dir), file=sys.stderr)
            print(str(e), file=sys.stderr)
            exit(1)

try:
    os.remove(args.touch_file)
except:
    pass

proc = subprocess.Popen(command, cwd=working_dir, stderr=subprocess.PIPE)
err = False
was_warn = False

for line in proc.stderr:
    line = line.decode('utf-8')

    if not line.strip() and was_warn:
        continue

    if 'RemovedInSphinx30Warning' in line:
        was_warn = True
        continue

    if 'warnings.warn' in line:
        was_warn = True
        continue

    if 'warning: Detected potential recursive class relation' in line:
        was_warn = True
        continue

    was_warn = False

    line = re.sub('WARNING:', 'warning:', line, flags=re.I)
    line = re.sub('ERROR:', 'error:', line, flags=re.I)

    m = re.match(r'^(warning:\s*|error:\s*)?([^:.]+[.][^:.]+)(:.*)', line)
    if m:
        line = (m.group(1) or '') + os.path.relpath(m.group(2), project_dir) + m.group(3)

    if args.werror and re.search('warning:', line):
        line = re.sub('warning:', 'error:', line)
        err = True
    elif 'error:' in line:
        err = True

    print(line, file=sys.stderr)

ret = proc.wait()

if ret > 0:
    exit(ret)
elif ret < 0:
    os.kill(os.getpid(), -ret)

if err:
    exit(127)

try:
    open(args.touch_file, 'w').close()
except:
    e = sys.exc_info()[1]
    print("error: unable to touch '{}' file".format(args.touch_file), file=sys.stderr)
    print(str(e), file=sys.stderr)
    exit(1)

exit(0)



import os

def WrapLauncher(env, tool, launcher):
    return '"{launcher}" "{tool}"'.format(
        launcher=launcher, tool=tool)

def WrapClangDb(env, tool, build_dir):
    if not tool.startswith('"'):
        tool = '"{}"'.format(tool)

    return (
        '{python_cmd} scripts/scons_helpers/clangdb.py '
        '"{root_dir}" "{build_dir}" {tool}'.format(
            python_cmd=env.GetPythonExecutable(),
            root_dir=env.Dir('#').path,
            build_dir=env.Dir(build_dir).path,
            tool=tool
        )
    )

def init(env):
    env.AddMethod(WrapLauncher, 'WrapLauncher')
    env.AddMethod(WrapClangDb, 'WrapClangDb')



import os.path
import platform
import re

def _fix_target(s):
    # on redhat and fedora, "system" and "os" parts are different
    # convert it to common form
    s = s.replace('-redhat-linux', '-linux-gnu')
    parts = s.split('-')
    if len(parts) == 3:
        # on some distros "system" part is omitted
        # convert it to common form
        parts = [parts[0]] + ['pc'] + parts[1:]
    elif len(parts) == 4:
        # on older config.guess versions "system" part defaulted to
        # "unknown" instead of "pc"; convert it to modern form
        if parts[1] == 'unknown':
            parts[1] = 'pc'
    s =  '-'.join(parts)
    return s

def ParseGitHead(env):
    try:
        with open('.git/HEAD') as hf:
            head_ref = hf.read().strip().split()[-1]
            pattern = re.compile(r'\b[0-9a-f]{40}\b')
            is_sha = re.match(pattern, head_ref)
            if is_sha:
                return head_ref[:10]
            else:
                with open('.git/{}'.format(head_ref)) as hrf:
                    return hrf.read().strip()[:10]
    except:
        return None

def ParseProjectVersion(env, path):
    return env.GetCommandOutput(
        '{python_cmd} scripts/scons_helpers/parse-version.py'.format(
            python_cmd=env.GetPythonExecutable()))

def ParseToolVersion(env, command):
    text = env.GetCommandOutput(command)
    if not text:
        return None

    m = re.search(r'(\b[0-9][0-9.]+\b)', text)
    if not m:
        return None

    return m.group(1)

def ParseCompilerVersion(env, compiler):
    def getverstr():
        try:
            version_formats = [
                r'(\b[0-9]+\.[0-9]+\.[0-9]+\b)',
                r'(\b[0-9]+\.[0-9]+\b)',
            ]

            full_text = env.GetCommandOutput(compiler + ' --version')

            for regex in version_formats:
                m = re.search(r'(?:LLVM|clang)\s+version\s+'+regex, full_text)
                if m:
                    return m.group(1)

            trunc_text = re.sub(r'\([^)]+\)', '', full_text)

            dump_text = env.GetCommandOutput(compiler + ' -dumpversion')

            for text in [dump_text, trunc_text, full_text]:
                for regex in version_formats:
                    m = re.search(regex, text)
                    if m:
                        return m.group(1)

            return None
        except:
            pass

    ver = getverstr()
    if ver:
        return tuple(map(int, ver.split('.')))
    else:
        return None

def ParseCompilerTarget(env, compiler):
    text = env.GetCommandOutput(compiler + ' -v -E -')
    if not text:
        return None

    for line in text.splitlines():
        m = re.search(r'\bTarget:\s*(\S+)', line)
        if m:
            s = m.group(1)
            s = _fix_target(s)
            return s

    return None

def ParseCompilerType(env, compiler):
    cmds = [compiler + ' --version', compiler + ' -v -E -']
    for cmd in cmds:
        text = env.GetCommandOutput(cmd)
        if 'clang version' in text:
            return 'clang'
        if 'gcc version' in text:
            return 'gcc'

    basename = os.path.basename(compiler)
    if 'clang' in basename:
        return 'clang'
    if 'gcc' in basename or 'g++' in basename:
        return 'gcc'
    if basename in ['cc', 'c++']:
        return 'cc'

    return None

def ParseCompilerDirectory(env, compiler):
    text = env.GetCommandOutput(compiler + ' --version')
    if not text:
        return None

    for line in text.splitlines():
        m = re.search(r'\bInstalledDir:\s*(.*)', line)
        if m:
            return m.group(1)

    return None

def ParseLinkDirs(env, linker):
    text = env.GetCommandOutput(linker + ' -print-search-dirs')
    if not text:
        return []

    for line in text.splitlines():
        m = re.search(r'^\s*libraries:\s*=(.*)$', line)
        if m:
            ret = []
            for libdir in m.group(1).split(':'):
                libdir = os.path.abspath(libdir)
                if os.path.isdir(libdir):
                    ret.append(libdir)
            return ret

    return []

def ParseConfigGuess(env, cmd):
    text = env.GetCommandOutput(cmd)
    if not text:
        return None

    if not re.match(r'^\S+-\S+$', text):
        return None

    text = _fix_target(text)
    return text

def ParseMacosHost(env, host, macos_platform, macos_arch):
    if not 'darwin' in host:
        return host

    arch_list = ParseList(env, macos_arch, ['arm64', 'x86_64'])
    if len(arch_list) > 1:
        for arch in ['arm64', 'x86_64']:
            host = host.replace(arch, 'universal')

    host = host.replace('-pc-apple-', '-apple-macos{}-'.format(
        ParseMacosPlatform(env, host, macos_platform)))

    return host

def ParseMacosPlatform(env, host, macos_platform):
    if not 'darwin' in host:
        return None

    if macos_platform:
        return macos_platform

    text = env.GetCommandOutput('sw_vers')
    if text:
        for line in text.splitlines():
            if 'ProductVersion' in line:
                try:
                    return '.'.join(line.split()[-1].split('.')[:2])
                except:
                    pass

    return None

def ParseMacosArch(env, host, macos_arch):
    if not 'darwin' in host:
        return []

    arch_list = ParseList(env, macos_arch, ['arm64', 'x86_64'])
    if arch_list:
        return arch_list

    if 'universal' in host:
        return ['arm64', 'x86_64']

    for arch in ['arm64', 'x86_64']:
        if arch in host:
            return [arch]

    if 'arm' in platform.processor():
        return ['arm64']
    else:
        return ['x86_64']

def ParseList(env, s, all):
    if not s:
        return []
    ret = []
    for name in s.split(','):
        if name == 'all':
            for other_name in all:
                if not other_name in ret and other_name != 'all':
                    ret.append(other_name)
        else:
            if not name in ret:
                ret.append(name)
    return ret

def init(env):
    env.AddMethod(ParseGitHead, 'ParseGitHead')
    env.AddMethod(ParseProjectVersion, 'ParseProjectVersion')
    env.AddMethod(ParseToolVersion, 'ParseToolVersion')
    env.AddMethod(ParseCompilerVersion, 'ParseCompilerVersion')
    env.AddMethod(ParseCompilerTarget, 'ParseCompilerTarget')
    env.AddMethod(ParseCompilerType, 'ParseCompilerType')
    env.AddMethod(ParseCompilerDirectory, 'ParseCompilerDirectory')
    env.AddMethod(ParseLinkDirs, 'ParseLinkDirs')
    env.AddMethod(ParseConfigGuess, 'ParseConfigGuess')
    env.AddMethod(ParseMacosHost, 'ParseMacosHost')
    env.AddMethod(ParseMacosPlatform, 'ParseMacosPlatform')
    env.AddMethod(ParseMacosArch, 'ParseMacosArch')
    env.AddMethod(ParseList, 'ParseList')



from __future__ import print_function

import argparse
import doctest
import fileinput
import fnmatch
import glob
import multiprocessing
import os
import os.path
import re
import shutil
import ssl
import subprocess
import sys
import tarfile
import textwrap

try:
    from shlex import quote
except:
    from pipes import quote

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

try:
    # workaround for SSL certificate error
    ssl._create_default_https_context = ssl._create_unverified_context
except:
    pass

DEV_NULL = open(os.devnull, 'w+')

Context = type(
    'Context', (),
    {
        field: '' for field in (
            'root_dir work_dir dist_dir log_file commit_file'
            ' pkg_dir pkg_src_dir pkg_bin_dir pkg_lib_dir pkg_inc_dir pkg_rpath_dir'
            ' pkg pkg_name pkg_ver pkg_deps'
            ' build host toolchain variant android_platform macos_platform macos_arch'
            ' prefer_cmake'
            ' env unparsed_env').split()
    })

#
# Helpers.
#

def changedir(ctx, dir_path):
    # format arguments
    dir_path = dir_path.format(ctx=ctx)

    with open(ctx.log_file, 'a+') as fp:
        print('>>> cd ' + dir_path, file=fp)

    os.chdir(dir_path)

def download(ctx, url, out):
    def _fetch_vendored(url, path):
        for subdir, _, _ in os.walk(ctx.dist_dir):
            distfile = os.path.join(subdir, os.path.basename(path))
            if os.path.exists(distfile):
                msg('[found vendored] {}', os.path.basename(distfile))
                with open(distfile, 'rb') as rp:
                    with open(path, 'wb') as wp:
                        wp.write(rp.read())
                        return
        raise

    def _fetch_urlopen(url, path):
        rp = urlopen(url)
        with open(path, 'wb') as wp:
            wp.write(rp.read())

    def _fetch_tool(url, path, tool, cmd):
        msg('[trying {}] {}', tool, url)
        with open(ctx.log_file, 'a+') as fp:
            print('>>> ' + cmd, file=fp)
        if os.system(cmd) != 0:
            raise

    def _fetch_wget(url, path):
        _fetch_tool(url, path, 'wget',
                    'wget "{}" --quiet -O "{}"'.format(url, path))

    def _fetch_curl(url, path):
        _fetch_tool(url, path, 'curl',
                    'curl -Ls "{}" -o "{}"'.format(url, path))

    # format arguments
    url = url.format(ctx=ctx)
    out = out.format(ctx=ctx)

    path_res = 'src/' + out
    path_tmp = 'tmp/' + out

    if os.path.exists(path_res):
        msg('[found downloaded] {}', out)
        return

    rmpath(path_res)
    rmpath(path_tmp)
    mkpath('tmp')

    error = None
    for fn in [_fetch_vendored, _fetch_urlopen, _fetch_curl, _fetch_wget]:
        try:
            fn(url, path_tmp)
            shutil.move(path_tmp, path_res)
            rm_emptydir('tmp')
            return
        except Exception as e:
            if fn == _fetch_vendored:
                msg('[download] {}', url)
            if fn == _fetch_urlopen:
                error = e

    die("can't download '{}': {}", url, error)

def unpack(ctx, filename, dirname):
    # format arguments
    filename = filename.format(ctx=ctx)
    dirname = dirname.format(ctx=ctx)

    dirname_res = 'src/' + dirname
    dirname_tmp = 'tmp/' + dirname

    if os.path.exists(dirname_res):
        msg('[found unpacked] {}', dirname)
        return

    msg('[unpack] {}', filename)

    rmpath(dirname_res)
    rmpath(dirname_tmp)
    mkpath('tmp')

    tar = tarfile.open('src/'+filename, 'r')
    tar.extractall('tmp')
    tar.close()

    shutil.move(dirname_tmp, dirname_res)
    rm_emptydir('tmp')

def execute(ctx, cmd, ignore_error=False, clear_env=False):
    msg('[execute] {}', cmd)

    with open(ctx.log_file, 'a+') as fp:
        print('>>> ' + cmd, file=fp)

    env = None
    if clear_env:
        env = {'HOME': os.environ['HOME'], 'PATH': os.environ['PATH']}

    code = subprocess.call('{} >>{} 2>&1'.format(cmd, ctx.log_file), shell=True, env=env)
    if code != 0:
        if ignore_error:
            with open(ctx.log_file, 'a+') as fp:
                print('command exited with status {}'.format(code), file=fp)
        else:
            exit(1)

def execute_make(ctx, cpu_count=None):
    if cpu_count is None:
        cpu_count = detect_cpu_count()

    cmd = ['make']
    if cpu_count:
        cmd += ['-j' + str(cpu_count)]

    execute(ctx, ' '.join(cmd))

def execute_cmake(ctx, src_dir, args=None):
    def _getvar(var, default):
        if var in ctx.env:
            return ctx.env[var]
        return '-'.join([s for s in [ctx.toolchain, default] if s])

    # format arguments
    src_dir = src_dir.format(ctx=ctx)

    if not args:
        args = []

    compiler = _getvar('CC', 'gcc')
    sysroot = find_sysroot(ctx.toolchain, compiler)

    need_sysroot = bool(sysroot)
    need_tools = True

    # cross-compiling for yocto linux
    if 'OE_CMAKE_TOOLCHAIN_FILE' in os.environ:
        need_tools = False

    # building for macOS
    if ctx.macos_platform:
        args += [
            '-DCMAKE_OSX_DEPLOYMENT_TARGET=' + quote(ctx.macos_platform),
        ]

        if ctx.macos_arch:
            args += [
                '-DCMAKE_OSX_ARCHITECTURES=' + quote(';'.join(ctx.macos_arch)),
            ]

    # cross-compiling for Android
    if ctx.android_platform:
        args += [
            '-DCMAKE_SYSTEM_NAME=Android',
            '-DANDROID_PLATFORM=android-' + ctx.android_platform,
        ]

        api = detect_android_api(compiler)
        abi = detect_android_abi(ctx.toolchain)

        toolchain_file = find_android_toolchain_file(compiler)

        if toolchain_file:
            need_sysroot = False
            need_tools = False
            args += [quote('-DCMAKE_TOOLCHAIN_FILE=' + toolchain_file)]
            if api:
                args += ['-DANDROID_NATIVE_API_LEVEL=' + api]
            if abi:
                args += ['-DANDROID_ABI=' + abi]
        else:
            sysroot = find_android_sysroot(compiler)
            need_sysroot = bool(sysroot)
            need_tools = True
            if api:
                args += ['-DCMAKE_SYSTEM_VERSION=' + api]
            if abi:
                args += ['-DCMAKE_ANDROID_ARCH_ABI=' + abi]

    if need_sysroot:
        args += [
            '-DCMAKE_FIND_ROOT_PATH=' + quote(sysroot),
            '-DCMAKE_SYSROOT='        + quote(sysroot),
        ]

    if need_tools:
        if not ctx.android_platform:
            args += [
                '-DCMAKE_C_COMPILER=' + quote(find_tool(compiler)),
            ]
        args += [
            '-DCMAKE_LINKER=' + quote(find_tool(_getvar('CCLD', 'gcc'))),
            '-DCMAKE_AR='     + quote(find_tool(_getvar('AR', 'ar'))),
            '-DCMAKE_RANLIB=' + quote(find_tool(_getvar('RANLIB', 'ranlib'))),
        ]

    if ctx.env.get('COMPILER_LAUNCHER', None):
        args += [
            '-DCMAKE_CXX_COMPILER_LAUNCHER=' + quote(ctx.env['COMPILER_LAUNCHER']),
            '-DCMAKE_C_COMPILER_LAUNCHER=' + quote(ctx.env['COMPILER_LAUNCHER']),
        ]

    args += [
        '-DCMAKE_POSITION_INDEPENDENT_CODE=ON',
    ]

    cc_flags = [
        '-fPIC', # -fPIC should be set explicitly in older cmake versions
    ]

    if ctx.variant == 'debug':
        cc_flags += [
            '-ggdb', # -ggdb is required for sanitizer backtrace
        ]
        args += [
            '-DCMAKE_BUILD_TYPE=Debug',
            '-DCMAKE_C_FLAGS_DEBUG:STRING=' + quote(' '.join(cc_flags)),
        ]
    else:
        args += [
            '-DCMAKE_BUILD_TYPE=Release',
            '-DCMAKE_C_FLAGS_RELEASE:STRING=' + quote(' '.join(cc_flags)),
        ]

    execute(ctx, 'cmake ' + src_dir + ' ' + ' '.join(args))

def execute_cmake_build(ctx):
    cmd = ['cmake', '--build', '.']

    # assume we're using -G"Unix Makefiles"
    cmd += ['--', 'VERBOSE=1']
    cpu_count = detect_cpu_count()
    if cpu_count:
        cmd += ['-j' + str(cpu_count)]

    execute(ctx, ' '.join(cmd))

def install_tree(ctx, src, dst, include=None, exclude=None):
    # format arguments
    src = src.format(ctx=ctx)
    dst = dst.format(ctx=ctx)

    msg('[install] {}', os.path.relpath(dst, ctx.root_dir))

    def _match_patterns(src, names):
        ignore_names = []
        for n in names:
            if os.path.isdir(os.path.join(src, n)):
                continue
            matched = False
            for m in include:
                if fnmatch.fnmatch(n, m):
                    matched = True
                    break
            if not matched:
                ignore_names.append(n)
        return set(ignore_names)

    if include:
        ignore_fn = _match_patterns
    elif exclude:
        ignore_fn = shutil.ignore_patterns(*exclude)
    else:
        ignore_fn = None

    mkpath(os.path.dirname(dst))
    rmpath(dst)
    shutil.copytree(src, dst, ignore=ignore_fn)

def install_files(ctx, src, dst):
    # format arguments
    src = src.format(ctx=ctx)
    dst = dst.format(ctx=ctx)

    msg('[install] {}', os.path.join(
        os.path.relpath(dst, ctx.root_dir),
        os.path.basename(src)))

    for f in glob.glob(src):
        mkpath(dst)
        shutil.copy(f, dst)

def subst_tree(ctx, dir_path, file_patterns, from_, to):
    def _match_path(path):
        try:
            with open(path) as fp:
                for line in fp:
                    if from_ in line:
                        return True
        except:
            pass

    # format arguments
    dir_path = dir_path.format(ctx=ctx)

    for filepat in file_patterns:
        for root, dirnames, filenames in os.walk(dir_path):
            for filename in fnmatch.filter(filenames, filepat):
                filepath = os.path.join(root, filename)
                if _match_path(filepath):
                    subst_files(ctx, filepath, from_, to)

def subst_files(ctx, file_path, from_, to):
    # format arguments
    file_path = file_path.format(ctx=ctx)

    msg('[patch] {}', file_path)

    for line in fileinput.input(file_path, inplace=True):
        print(line.replace(from_, to), end='')

def apply_patch(ctx, dir_path, patch_url, patch_name):
    # format arguments
    dir_path = dir_path.format(ctx=ctx)

    if subprocess.call(
        'patch --version', stdout=DEV_NULL, stderr=DEV_NULL, shell=True) != 0:
        return

    download(ctx, patch_url, patch_name)
    execute(ctx, 'patch -p1 -N -d {} -i {}'.format(
        dir_path,
        '../' + patch_name), ignore_error=True)

def format_vars(ctx, disable_launcher=False):
    ret = []
    for k, v in ctx.env.items():
        if k == 'COMPILER_LAUNCHER':
            continue
        elif k in ['CC', 'CXX'] and not disable_launcher:
            if ctx.env.get('COMPILER_LAUNCHER', None):
                ret.append(quote('{}={} {}'.format(k, ctx.env['COMPILER_LAUNCHER'], v)))
            else:
                ret.append(quote('{}={}'.format(k, v)))
        else:
            ret.append(quote('{}={}'.format(k, v)))
    return ' '.join(ret)

def format_flags(ctx, cflags='', ldflags='', pthread=False):
    inc_dirs = []
    lib_dirs = []
    rpath_dirs = []

    for dep in ctx.pkg_deps:
        inc_dirs += [os.path.join(ctx.work_dir, dep, 'include')]
        lib_dirs += [os.path.join(ctx.work_dir, dep, 'lib')]
        rpath_dirs += [os.path.join(ctx.work_dir, dep, 'rpath')]

    cflags = ([cflags] if cflags else []) + ['-I' + path for path in inc_dirs]
    ldflags = ['-L' + path for path in lib_dirs] + ([ldflags] if ldflags else [])

    is_gnu = detect_compiler_family(ctx.env, ctx.toolchain, 'gcc')
    is_clang = detect_compiler_family(ctx.env, ctx.toolchain, 'clang')

    if ctx.variant == 'debug':
        if is_gnu or is_clang:
            cflags += ['-ggdb']
        else:
            cflags += ['-g']
    elif ctx.variant == 'release':
        cflags += ['-O2']

    if pthread and not ctx.android_platform:
        if is_gnu or is_clang:
            cflags += ['-pthread']
        if is_gnu:
            ldflags += ['-pthread']
        else:
            ldflags += ['-lpthread']

    if is_gnu:
        ldflags += ['-Wl,-rpath-link=' + path for path in rpath_dirs]

    if ctx.macos_platform:
        cflags += ['-mmacosx-version-min=' + ctx.macos_platform]
        ldflags += ['-mmacosx-version-min=' + ctx.macos_platform]

        for arch in ctx.macos_arch:
            cflags += ['-arch', arch]
            ldflags += ['-arch', arch]

    return ' '.join([
        'CXXFLAGS=' + quote(' '.join(cflags)),
        'CFLAGS='   + quote(' '.join(cflags)),
        'LDFLAGS='  + quote(' '.join(ldflags)),
    ])

# https://mesonbuild.com/Reference-tables.html
def generate_meson_crossfile(ctx, pc_dir, cross_file):
    def _meson_string(s):
        m = {
            "\\": "\\\\",
            "'":  "\\'",
            "\a": "\\a",
            "\b": "\\b",
            "\f": "\\f",
            "\n": "\\n",
            "\r": "\\r",
            "\t": "\\t",
            "\v": "\\v",
        }
        for k, v in m.items():
            s = s.replace(k, v)
        return "'" + s + "'"

    def _meson_list(l):
        return '[{}]'.format(', '.join(map(_meson_string, l)))

    def _meson_compiler(compiler, launcher):
        if launcher:
            return _meson_list([launcher, compiler])
        else:
            return _meson_string(compiler)

    msg('[generate] {}', cross_file)

    pkg_config = None
    if 'PKG_CONFIG' in ctx.env:
        pkg_config = ctx.env['PKG_CONFIG']
    elif which('pkg-config'):
        pkg_config = 'pkg-config'

    cflags = []
    ldflags = []

    for dep in ctx.pkg_deps:
        cflags.append('-I' + os.path.join(ctx.work_dir, dep, 'include'))
        ldflags.append('-L' + os.path.join(ctx.work_dir, dep, 'lib'))

    if ctx.toolchain:
        system = 'linux'
        cpu_family = 'x86_64'
        cpu = 'x86_64'
        endian = 'little'

        system_list = ('android cygwin darwin dragonfly emscripten freebsd haiku netbsd'+
                   ' openbsd windows sunos').split()

        cpu_list = ('aarch64 alpha arc arm avr c2000 csky dspic e2k ft32 ia64'+
                    ' loongarch64 m68k microblaze mips mips64 msp430 parisc pic24'+
                    ' ppc ppc64 riscv32 riscv64 rl78 rx s390 s390x sh4 sparc sparc64'+
                    ' wasm32 wasm64 x86 x86_64').split()

        for s in reversed(sorted(system_list, key=len)):
            if s in ctx.toolchain:
                system = s
                break

        for s in reversed(sorted(cpu_list, key=len)):
            if s in ctx.toolchain:
                cpu_family = s

                if s == 'aarch64':
                    cpu = 'armv8'
                elif s == 'arm':
                    cpu = 'armv5'
                elif s == 'x86':
                    cpu = 'i686'
                else:
                    cpu = s

                if s+'le' in ctx.toolchain or s+'el' in ctx.toolchain:
                    endian = 'little'
                elif s+'be' in ctx.toolchain or s+'eb' in ctx.toolchain:
                    endian = 'big'
                elif s in 'wasm32 wasm64 x86 x86_64'.split():
                    endian = 'little'
                else:
                    endian = 'big'

                break

    with open(cross_file, 'w') as fp:
        fp.write(textwrap.dedent("""\
                [binaries]
                c = {cc}
                cpp = {cxx}
                ar = {ar}
        """).format(
            cc=_meson_compiler(ctx.env['CC'],
                ctx.env.get('COMPILER_LAUNCHER', None)),
            cxx=_meson_compiler(ctx.env['CXX'],
                ctx.env.get('COMPILER_LAUNCHER', None)),
            ar=_meson_string(ctx.env['AR'])))

        if pkg_config:
            fp.write(textwrap.dedent("""\
                pkgconfig = {pkg_config}
            """).format(
                pkg_config=_meson_string(pkg_config)))

        if pc_dir:
            fp.write(textwrap.dedent("""\

                [built-in options]
                pkg_config_path = {pc_dir}
            """).format(
                pc_dir=_meson_string(os.path.abspath(pc_dir))))

        if cflags or ldflags:
            fp.write(textwrap.dedent("""\

                [properties]
                c_args = {cflags}
                c_link_args = {ldflags}
                cpp_args = {cflags}
                cpp_link_args = {ldflags}
            """).format(
                cflags=_meson_list(cflags),
                ldflags=_meson_list(ldflags)))

        if ctx.toolchain:
            fp.write(textwrap.dedent("""\

                [host_machine]
                system = {system}
                cpu_family = {cpu_family}
                cpu = {cpu}
                endian = {endian}
            """).format(
                system=_meson_string(system),
                cpu_family=_meson_string(cpu_family),
                cpu=_meson_string(cpu),
                endian=_meson_string(endian)))

# Generate temporary .pc files for pkg-config.
def generate_pc_files(ctx, pc_dir):
    for dep in ctx.pkg_deps:
        name, ver = parse_dep(dep)
        pc_file = os.path.join(pc_dir, name) + '.pc'

        with open(pc_file, 'w') as fp:
            msg('[generate] {}', pc_file)

            cflags = '-I' + os.path.join(ctx.work_dir, dep, 'include')
            ldflags = '-L' + os.path.join(ctx.work_dir, dep, 'lib')

            for lib in glob.glob(os.path.join(ctx.work_dir, dep, 'lib', 'lib*')):
                lib = os.path.basename(lib)
                lib = re.sub('^lib', '', lib)
                lib = re.sub(r'\.[a-z]+$', '', lib)

                ldflags += ' -l' + lib

            fp.write(textwrap.dedent("""\
                Name: {name}
                Description: {name}
                Version: {ver}
                Cflags: {cflags}
                Libs: {ldflags}
            """).format(
                name=name,
                ver=ver,
                cflags=cflags,
                ldflags=ldflags))

def find_tool(tool):
    if '/' in tool:
        return tool

    path = which(tool)
    if not path:
        return tool

    return path

def find_sysroot(toolchain, compiler):
    if not toolchain:
        return ''

    if not compiler:
        compiler = '{}-gcc'.format(toolchain)

    try:
        sysroot = run_command([compiler, '-print-sysroot'])
        if os.path.isdir(sysroot):
            return sysroot
    except:
        pass

    return None

def find_android_sysroot(compiler):
    compiler_exe = which(compiler)
    if compiler_exe:
        return find_file_upwards(compiler_exe, 'sysroot')

def find_android_toolchain_file(compiler):
    compiler_exe = which(compiler)
    if compiler_exe:
        return find_file_upwards(compiler_exe, 'build/cmake/android.toolchain.cmake')

def detect_android_abi(toolchain):
    try:
        arch = toolchain.split('-')[0]
    except:
        return
    if arch == 'arm' or arch == 'armv7a':
        return 'armeabi-v7a'
    if arch == 'aarch64':
        return 'arm64-v8a'
    if arch == 'i686':
        return 'x86'
    if arch == 'x86_64':
        return 'x86_64'
    return arch

def detect_android_api(compiler):
    try:
        for line in run_command([compiler, '-dM', '-E', '-']).splitlines():
            m = re.search(r'__ANDROID_API__\s+(\d+)', line)
            if m:
                return m.group(1)
    except:
        pass

def detect_compiler_family(env, toolchain, family):
    if family == 'gcc':
        keys = ['GNU', 'gnu', 'gcc', 'g++']
    elif family == 'clang':
        keys = ['clang']

    def _check_tool(toolchain, tool):
        if toolchain:
            tool = '{}-{}'.format(toolchain, tool)
        try:
            out = run_command([tool, '-v'],
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            for k in keys:
                if k in out:
                    return True
        except:
            pass
        return False

    for var in ['CC', 'CCLD', 'CXX', 'CXXLD']:
        if var in env:
            if not _check_tool('', env[var]):
                return False

    if not 'gnu' in toolchain:
        if 'CC' not in env:
            for tool in ['cc', 'gcc', 'clang']:
                if _check_tool(toolchain, tool):
                    break
            else:
                return False

        if 'CCLD' not in env:
            for tool in ['ld', 'gcc', 'clang']:
                if _check_tool(toolchain, tool):
                    break
            else:
                return False

        if 'CXX' not in env or 'CXXLD' not in env:
            for tool in ['g++', 'clang++']:
                if _check_tool(toolchain, tool):
                    break
            else:
                return False

    return True

# Guess platform argument for OpenSSL's Configure script basing on toolchain string.
# (see `./Configure LIST` for a list of all platforms and their format)
# See also openssl/Configurations/[0-9]*.conf
def detect_openssl_platform(toolchain):
    """
    >>> detect_openssl_platform('x86_64-pc-linux-gnu')
    'linux-x86_64'
    >>> detect_openssl_platform('aarch64-linux-gnu')
    'linux-aarch64'
    >>> detect_openssl_platform('mips-openwrt-linux-uclibc')
    'linux-mips32'
    >>> detect_openssl_platform('arm-linux-gnueabihf')
    'linux-generic32'
    >>> detect_openssl_platform('arm-bcm2708hardfp-linux-gnueabi')
    'linux-generic32'
    >>> detect_openssl_platform('armv7a-linux-androideabi33')
    'android-arm'
    >>> detect_openssl_platform('mips64-linux-musl')
    'linux64-mips64'
    >>> detect_openssl_platform('arm-linux-androideabi')
    'android-arm'
    >>> detect_openssl_platform('aarch64-linux-android')
    'android-arm64'
    >>> detect_openssl_platform('aarch64-linux-android29')
    'android-arm64'
    >>> detect_openssl_platform('i686-linux-android')
    'android-x86'
    >>> detect_openssl_platform('x86_64-linux-android')
    'android-x86_64'
    >>> detect_openssl_platform('i686-pc-linux-gnu')
    'linux-x86'

    # other edge cases
    >>> detect_openssl_platform('mingw-w64-x86_64')
    'mingw64'
    >>> detect_openssl_platform('x86_64-pc-cygwin')
    'Cygwin-x86_64'
    >>> detect_openssl_platform('')
    'cc'
    """

    # we assume that:
    # - 'android' has a higher priority than 'linux'
    # - if toolchain contains '64' and 'android' => we have android64 on 64-bit CPU
    # - if toolchain contains '64' and 'linux' => we have linux64 on 64-bit CPU
    # - openssl architectures 'aout', 'elf', 'latomic' are meaningless (they cannot be found in
    #   toolchain)
    # - compiler suffixes '-cc', '-gcc', '-clang' are meaningless too
    # - aix* OSes are never used :)
    # - 'android64' are currently just aliases for 'android' architectures;
    #   'android64-aarch64' is the same as 'android-arm64'

    bitness = 64 if '64' in toolchain else 32

    ### generic detection

    # XXX: this is a dirty parser based on very simple heuristics and verified on a
    # limited set of platforms

    # Set of platforms formatted as (OS, [arch1, ...]).
    #
    # `arch` is an architecture that openssl accepts for a given `os`. arch in a toolchain may
    # be the same but called differently, e.g. 'x86' vs 'i686'. To handle such cases,
    # `arch_aliases` are used (see below): if `arch` is not in toolchain, we try to search all
    # aliases (one by one) before moving to the next arch.
    #
    # Some tunings:
    # - 'android-armeabi' is an alias of 'android-arm' => removed 'armeabi'
    # - 'linux-armv4' fails on android toolchain; we use 'linux-generic32' => removed 'armv4'
    # - 'android64' is an alias for 'android' => removed 'android64'
    # - 'generic64' and 'generic32' never occur in toolchains => removed them
    platforms = [
        # order is significant here as we search substrings in a whole toolchain string!
        # [(os,      [arch, ...]), ...]
        ('BSD',      ['aarch64', 'ia64', 'riscv64',
                      'sparc64', 'sparcv8', 'x86_64', 'x86']),
        ('Cygwin',   ['i386', 'i486', 'i586', 'i686', 'x86_64', 'x86']),
        ('android',  ['arm64', 'arm', 'mips64', 'mips', 'x86_64', 'x86']),
        ('darwin64', ['arm64', 'ppc', 'x86_64']),
        ('darwin',   ['i386', 'ppc']),
        ('linux64',  ['loongarch64', 'mips64', 'riscv64', 's390x', 'sparcv9']),
        ('linux32',  ['s390x']),
        ('linux',    ['aarch64', 'arm64ilp32', 'ia64', 'mips64', 'mips32', 'ppc64le',
                      'ppc64', 'ppc', 'sparcv8', 'sparcv9', 'x86_64', 'x86', 'x32']),
        ('mingw64',  []),
        ('mingw',    []),
    ]
    arch_aliases = {
        'arm64':     ['aarch64'],
        'mips32':    ['mips'],
        'x86':       ['i386', 'i486', 'i586', 'i686'],
    }

    #~print('0: ', toolchain)
    for ossl_os, ossl_archs in platforms:
        os = ossl_os.lower()
        # when 64 bit, we explicitly prefer 'linux64' and similar (assuming libc and other
        # libs to be 64 bit)
        if os in toolchain or (bitness == 64 and os.endswith('64') and os[:-2] in toolchain):
            for arch in ossl_archs:
                #~print('1: ', os, arch)
                if arch in toolchain:
                    return '-'.join([ossl_os, arch])
                # or any of the aliases in toolchain
                for alias in arch_aliases.get(arch, []):
                    if alias in toolchain:
                        return '-'.join([ossl_os, arch])

            # ignore 'android64' without generality restriction
            if ossl_os in ['mingw', 'mingw64', 'Cygwin']:
                return ossl_os

            # on 'linux' and 'BSD', unless arch was found, return '...-generic{32,64}'
            if ossl_os in ['BSD', 'linux']:
                return '-'.join([os, 'generic64' if bitness == 64 else 'generic32'])
    return 'cc' # just a fallback

def detect_cpu_count():
    try:
        return len(os.sched_getaffinity(0))
    except:
        pass

    try:
        return multiprocessing.cpu_count()
    except:
        pass

    return None

def parse_env(unparsed_vars):
    env = dict()
    for e in unparsed_vars:
        k, v = e.split('=', 1)
        env[k] = v
    return env

def parse_dep(unparsed_dep):
    m = re.match('^(.*?)-([0-9][a-z0-9.-]+)$', unparsed_dep)
    if not m:
        die("can't determine version of '{}'", unparsed_dep)
    return m.group(1), m.group(2)

def parse_ver(unparsed_ver):
    try:
        return tuple(map(int, unparsed_ver.split('.')))
    except:
        die("can't parse version '{}'", unparsed_ver)

def run_command(cmd, stdin=DEV_NULL, stdout=subprocess.PIPE, stderr=DEV_NULL):
    proc = subprocess.Popen(
        cmd,
        shell=not isinstance(cmd, list),
        stdin=stdin, stdout=stdout, stderr=stderr)

    out = proc.stdout.read()

    try:
        out = out.decode()
    except:
        pass
    try:
        out = str(out)
    except:
        pass
    try:
        out = out.strip()
    except:
        pass
    return out

def find_file_upwards(path, search_file):
    while True:
        parent = os.path.dirname(path)
        if parent == path:
            break # root
        path = parent
        child_path = os.path.join(path, search_file)
        if os.path.exists(child_path):
            return child_path

def which(tool):
    try:
        path = shutil.which(tool)
        if path:
            return path
    except:
        pass

    out = run_command(['which', tool])
    if out:
        return out

def mkpath(path):
    try:
        os.makedirs(path)
    except:
        pass

def rmpath(path):
    try:
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)
    except:
        pass

def rm_emptydir(path):
    try:
        os.rmdir(path)
    except:
        pass

def touch(path):
    open(path, 'w').close()

def msg(text, *args):
    if args:
        text = text.format(*args)
    print(text)

def die(text, *args):
    if args:
        text = text.format(*args)
    print('error: ' + text, file=sys.stderr)
    exit(1)

#
# Parse command-line options.
#

if '--self-test' in sys.argv:
    failures, total = doctest.testmod(report=True)

    if failures == 0:
        print('OK: {} tests passed'.format(total))
        exit(0)
    else:
        exit(1)

parser = argparse.ArgumentParser(description='build third-party packages')

parser.add_argument('--self-test', action='store_true',
                    help='run doctests and exit')

parser.add_argument('--root-dir', dest='root_dir', type=str, required=True,
                    help='path to project root')

parser.add_argument('--work-dir', dest='work_dir', type=str, required=True,
                    help='path to working directory')

parser.add_argument('--dist-dir', dest='dist_dir', type=str, required=True,
                    help='path to vendored distfiles directory')

parser.add_argument('--build', dest='build', type=str, required=True,
                    help='system when package is built (e.g. x86_64-pc-linux-gnu)')

parser.add_argument('--host', dest='host', type=str, required=True,
                    help='system when package will run (e.g. aarch64-linux-gnu)')

parser.add_argument('--toolchain', dest='toolchain', type=str,
                    help='toolchain prefix (e.g. aarch64-linux-gnu)')

parser.add_argument('--variant', dest='variant', type=str, required=True,
                    help='build variant', choices=['debug', 'release'])

parser.add_argument('--package', dest='package', type=str, required=True,
                    help='package name and version (e.g. openssl-3.0.7-rc1)')

parser.add_argument('--deps', dest='deps', type=str, nargs='*',
                    help='package dependencies (should be built previously)')

parser.add_argument('--vars', dest='vars', type=str, nargs='*',
                    help='environment variables (e.g. CC=gcc CXX=g++ ...)')

parser.add_argument('--android-platform', dest='android_platform', type=str,
                    help='android target platform version')

parser.add_argument('--macos-platform', dest='macos_platform', type=str,
                    help='macos target platform version')

parser.add_argument('--macos-arch', dest='macos_arch', type=str, nargs='*',
                    help='macos target architecture(s)')

args = parser.parse_args()

#
# Setup context.
#

ctx = Context()

ctx.pkg = args.package
ctx.pkg_name, ctx.pkg_ver = parse_dep(args.package)
ctx.pkg_deps = args.deps or []

ctx.build = args.build
ctx.host = args.host
ctx.toolchain = args.toolchain or ''
ctx.variant = args.variant
ctx.android_platform = args.android_platform
ctx.macos_platform = args.macos_platform
ctx.macos_arch = args.macos_arch

ctx.unparsed_env = args.vars or []
ctx.env = parse_env(args.vars)

os.chdir(args.root_dir)

ctx.root_dir = os.path.abspath('.')
ctx.work_dir = os.path.abspath(args.work_dir)
ctx.dist_dir = os.path.abspath(args.dist_dir)

ctx.pkg_dir = os.path.join(ctx.work_dir, ctx.pkg)

ctx.pkg_src_dir = os.path.join(ctx.pkg_dir, 'src')
ctx.pkg_bin_dir = os.path.join(ctx.pkg_dir, 'bin')
ctx.pkg_lib_dir = os.path.join(ctx.pkg_dir, 'lib')
ctx.pkg_inc_dir = os.path.join(ctx.pkg_dir, 'include')
ctx.pkg_rpath_dir = os.path.join(ctx.pkg_dir, 'rpath')

ctx.log_file = os.path.join(ctx.pkg_dir, 'build.log')
ctx.commit_file = os.path.join(ctx.pkg_dir, 'commit')

ctx.prefer_cmake = bool(args.macos_platform or args.android_platform)

#
# Build package.
#

rmpath(ctx.commit_file)
mkpath(ctx.pkg_src_dir)

changedir(ctx, ctx.pkg_dir)

if ctx.pkg_name == 'libuv':
    download(
        ctx,
        'https://dist.libuv.org/dist/v{ctx.pkg_ver}/libuv-v{ctx.pkg_ver}.tar.gz',
        'libuv-v{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'libuv-v{ctx.pkg_ver}.tar.gz',
        'libuv-v{ctx.pkg_ver}')
    changedir(ctx, 'src/libuv-v{ctx.pkg_ver}')
    if ctx.prefer_cmake:
        mkpath('build')
        changedir(ctx, 'build')
        execute_cmake(ctx, '..', args=[
            '-DLIBUV_BUILD_TESTS=OFF',
            ])
        execute_cmake_build(ctx)
        if os.path.isfile('libuv_a.a'):
            shutil.copy('libuv_a.a', '../libuv.a')
        else:
            shutil.copy('libuv.a', '../libuv.a')
        changedir(ctx, '..')
    else:
        subst_files(ctx, 'src/unix/core.c', ' dup3', ' uv__dup3')
        execute(ctx, './autogen.sh')
        execute(ctx, './configure --host={host} {vars} {flags} {opts}'.format(
            host=ctx.toolchain,
            vars=format_vars(ctx),
            flags=format_flags(ctx),
            opts=' '.join([
                '--with-pic',
                '--enable-static',
            ])))
        execute_make(ctx)
        shutil.copy('.libs/libuv.a', 'libuv.a')
    install_tree(ctx, 'include', ctx.pkg_inc_dir)
    install_files(ctx, 'libuv.a', ctx.pkg_lib_dir)
elif ctx.pkg_name == 'libatomic_ops':
    download(
        ctx,
        'https://github.com/ivmai/libatomic_ops/releases/download/'
            'v{ctx.pkg_ver}/libatomic_ops-{ctx.pkg_ver}.tar.gz',
        'libatomic_ops-{ctx.pkg_ver}.tar.gz')
    unpack(ctx,
           'libatomic_ops-{ctx.pkg_ver}.tar.gz',
           'libatomic_ops-{ctx.pkg_ver}')
    changedir(ctx, 'src/libatomic_ops-{ctx.pkg_ver}')
    execute(ctx, './configure --host={host} {vars} {flags} {opts}'.format(
        host=ctx.toolchain,
        vars=format_vars(ctx),
        flags=format_flags(ctx, cflags='-fPIC'),
        opts=' '.join([
            '--disable-docs',
            '--disable-shared',
            '--enable-static',
           ])))
    execute_make(ctx)
    install_tree(ctx, 'src', ctx.pkg_inc_dir, include=['*.h'])
    install_files(ctx, 'src/.libs/libatomic_ops.a', ctx.pkg_lib_dir)
elif ctx.pkg_name == 'libunwind':
    download(
        ctx,
        'http://download.savannah.nongnu.org/releases/libunwind/'
            'libunwind-{ctx.pkg_ver}.tar.gz',
        'libunwind-{ctx.pkg_ver}.tar.gz')
    unpack(ctx,
           'libunwind-{ctx.pkg_ver}.tar.gz',
           'libunwind-{ctx.pkg_ver}')
    changedir(ctx, 'src/libunwind-{ctx.pkg_ver}')
    execute(ctx, './configure --host={host} {vars} {flags} {opts}'.format(
        host=ctx.toolchain,
        vars=format_vars(ctx),
        flags=format_flags(ctx, cflags='-fcommon -fPIC'),
        opts=' '.join([
            '--disable-coredump',
            '--disable-minidebuginfo',
            '--disable-ptrace',
            '--disable-setjmp',
            '--disable-shared',
            '--enable-static',
           ])))
    execute_make(ctx)
    install_files(ctx, 'include/*.h', ctx.pkg_inc_dir)
    install_files(ctx, 'src/.libs/libunwind.a', ctx.pkg_lib_dir)
elif ctx.pkg_name == 'openfec':
    if ctx.variant == 'debug':
        setattr(ctx, 'res_dir', 'bin/Debug')
    else:
        setattr(ctx, 'res_dir', 'bin/Release')
    download(
        ctx,
        'https://github.com/roc-streaming/openfec/archive/v{ctx.pkg_ver}.tar.gz',
        'openfec_v{ctx.pkg_ver}.tar.gz')
    unpack(ctx,
           'openfec_v{ctx.pkg_ver}.tar.gz',
           'openfec-{ctx.pkg_ver}')
    changedir(ctx, 'src/openfec-{ctx.pkg_ver}')
    mkpath('build')
    changedir(ctx, 'build')
    execute_cmake(ctx, '..', args=[
        '-DBUILD_STATIC_LIBS=ON',
        '-DDEBUG:STRING=%s' % ('ON' if ctx.variant == 'debug' else 'OFF'),
        ])
    execute_cmake_build(ctx)
    changedir(ctx, '..')
    install_tree(ctx, 'src', ctx.pkg_inc_dir, include=['*.h'])
    install_files(ctx, '{ctx.res_dir}/libopenfec.a', ctx.pkg_lib_dir)
elif ctx.pkg_name == 'openssl':
    download(
        ctx,
        'https://www.openssl.org/source/openssl-{ctx.pkg_ver}.tar.gz',
        'openssl-{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'openssl-{ctx.pkg_ver}.tar.gz',
        'openssl-{ctx.pkg_ver}')
    changedir(ctx, 'src/openssl-{ctx.pkg_ver}')
    # see https://github.com/openssl/openssl/blob/master/INSTALL.md#configuration-options
    execute(ctx, '{vars} {flags} ./Configure {platform} {variant} {options}'.format(
        vars=format_vars(ctx, disable_launcher=True),
        flags=format_flags(ctx),
        platform=detect_openssl_platform(ctx.host),
        variant='--debug' if ctx.variant == 'debug' else '--release',
        options=' '.join([
            'no-asan',
            'no-buildtest-c++',
            'no-external-tests',
            'no-fuzz-afl',
            'no-fuzz-libfuzzer',
            'no-shared',
            'no-tests',
            'no-ubsan',
            'no-ui-console',
            'no-unit-test',
        ])))
    execute_make(ctx)
    install_tree(ctx, 'include', ctx.pkg_inc_dir)
    install_files(ctx, 'libssl.a', ctx.pkg_lib_dir)
    install_files(ctx, 'libcrypto.a', ctx.pkg_lib_dir)
elif ctx.pkg_name == 'speexdsp':
    if ctx.pkg_ver.split('.', 1) > ['1', '2'] and (
            not re.match('^1.2[a-z]', ctx.pkg_ver) or ctx.pkg_ver == '1.2rc3'):
        setattr(ctx, 'pkg_repo', 'speexdsp')
    else:
        setattr(ctx, 'pkg_repo', 'speex')
    download(
        ctx,
        'http://downloads.xiph.org/releases/speex/{ctx.pkg_repo}-{ctx.pkg_ver}.tar.gz',
        '{ctx.pkg_repo}-{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        '{ctx.pkg_repo}-{ctx.pkg_ver}.tar.gz',
        '{ctx.pkg_repo}-{ctx.pkg_ver}')
    changedir(ctx, 'src/{ctx.pkg_repo}-{ctx.pkg_ver}')
    execute(ctx, './configure --host={host} {vars} {flags} {opts}'.format(
        host=ctx.toolchain,
        vars=format_vars(ctx),
        flags=format_flags(ctx, cflags='-fPIC'),
        opts=' '.join([
            '--disable-examples',
            '--disable-shared',
            '--enable-static',
           ])))
    execute_make(ctx)
    install_tree(ctx, 'include', ctx.pkg_inc_dir)
    install_files(ctx, 'lib{ctx.pkg_repo}/.libs/libspeexdsp.a', ctx.pkg_lib_dir)
elif ctx.pkg_name == 'sndfile':
    download(
        ctx,
        'http://www.mega-nerd.com/libsndfile/files/libsndfile-{ctx.pkg_ver}.tar.gz',
        'libsndfile-{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'libsndfile-{ctx.pkg_ver}.tar.gz',
        'libsndfile-{ctx.pkg_ver}')
    changedir(ctx, 'src/libsndfile-{ctx.pkg_ver}')
    execute(ctx, '{configure} --host={host} {vars} {flags} {opts}'.format(
        configure=' '.join(filter(None, [
            # workaround for outdated config.sub
            'ac_cv_host=%s' % ctx.toolchain if ctx.toolchain else '',
            # configure
            './configure',
        ])),
        host=ctx.toolchain,
        vars=format_vars(ctx),
        # explicitly enable -pthread because libtool doesn't add it on some platforms
        flags=format_flags(ctx, cflags='-fPIC', pthread=True),
        opts=' '.join([
            '--disable-external-libs',
            '--disable-shared',
            '--enable-static',
        ])))
    execute_make(ctx)
    install_files(ctx, 'src/sndfile.h', ctx.pkg_inc_dir)
    install_files(ctx, 'src/.libs/libsndfile.a', ctx.pkg_lib_dir)
elif ctx.pkg_name == 'sox':
    download(
        ctx,
        'https://downloads.sourceforge.net/project/sox/sox/'
            '{ctx.pkg_ver}/sox-{ctx.pkg_ver}.tar.gz',
        'sox-{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'sox-{ctx.pkg_ver}.tar.gz',
        'sox-{ctx.pkg_ver}')
    changedir(ctx, 'src/sox-{ctx.pkg_ver}')
    execute(ctx, './configure --host={host} {vars} {flags} {opts}'.format(
        host=ctx.toolchain,
        vars=format_vars(ctx),
        flags=format_flags(ctx),
        opts=' '.join([
            '--disable-openmp',
            '--disable-shared',
            '--enable-static',
            '--with-amrnb=no',
            '--with-amrwb=no',
            '--with-ao=no',
            '--with-flac=no',
            '--with-gsm=no',
            '--with-lpc10=no',
            '--with-mp3=no',
            '--with-oggvorbis=no',
            '--with-opus=no',
            '--with-pulseaudio=no',
            '--with-sndfile=no',
            '--with-sndio=no',
            '--with-wavpack=no',
            '--without-ao',
            '--without-id3tag',
            '--without-ladspa',
            '--without-lame',
            '--without-libltdl',
            '--without-mad',
            '--without-magic',
            '--without-opus',
            '--without-png',
            '--without-twolame',
        ])))
    execute_make(ctx)
    install_files(ctx, 'src/sox.h', ctx.pkg_inc_dir)
    install_files(ctx, 'src/.libs/libsox.a', ctx.pkg_lib_dir)
elif ctx.pkg_name == 'alsa':
    download(
        ctx,
        'https://www.alsa-project.org/files/pub/lib/alsa-lib-{ctx.pkg_ver}.tar.bz2',
        'alsa-lib-{ctx.pkg_ver}.tar.bz2')
    unpack(ctx,
           'alsa-lib-{ctx.pkg_ver}.tar.bz2',
           'alsa-lib-{ctx.pkg_ver}')
    changedir(ctx, 'src/alsa-lib-{ctx.pkg_ver}')
    execute(ctx, './configure --host={host} {vars} {opts}'.format(
        host=ctx.toolchain,
        vars=format_vars(ctx),
        opts=' '.join([
            '--disable-python',
            '--disable-static',
            '--enable-shared',
        ])))
    execute_make(ctx)
    install_tree(ctx, 'include/alsa',
            os.path.join(ctx.pkg_inc_dir, 'alsa'),
            exclude=['alsa'])
    install_files(ctx, 'src/.libs/libasound.so', ctx.pkg_lib_dir)
    install_files(ctx, 'src/.libs/libasound.so.*', ctx.pkg_rpath_dir)
elif ctx.pkg_name == 'pulseaudio':
    download(
        ctx,
        'https://freedesktop.org/software/pulseaudio/releases/'
            'pulseaudio-{ctx.pkg_ver}.tar.gz',
        'pulseaudio-{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'pulseaudio-{ctx.pkg_ver}.tar.gz',
        'pulseaudio-{ctx.pkg_ver}')
    pa_ver = parse_ver(ctx.pkg_ver)
    if (8, 99, 1) <= pa_ver < (11, 99, 1):
        apply_patch(
            ctx,
            'src/pulseaudio-{ctx.pkg_ver}',
            'https://bugs.freedesktop.org/attachment.cgi?id=136927',
            '0001-memfd-wrappers-only-define-memfd_create-if-not-alrea.patch')
    if pa_ver < (12, 99, 1):
        subst_tree(ctx, 'src/pulseaudio-{ctx.pkg_ver}', ['*.h', '*.c'],
                   from_='#include <asoundlib.h>',
                   to='#include <alsa/asoundlib.h>')
    changedir(ctx, 'src/pulseaudio-{ctx.pkg_ver}')
    if pa_ver < (14, 99, 1):
        # workaround for "missing acolocal-1.15" and "missing automake-1.15" errors
        # on some systems; since we're not modifying any autotools stuff, it's safe
        # to replace corresponding commands with "true" command
        if os.path.exists('Makefile.in'):
            subst_files(ctx, 'Makefile.in', '@ACLOCAL@', 'true')
            subst_files(ctx, 'Makefile.in', '@AUTOMAKE@', 'true')
        execute(ctx, './configure --host={host} {vars} {flags} {deps} {opts}'.format(
            host=ctx.toolchain,
            vars=format_vars(ctx),
            flags=format_flags(ctx, cflags='-w -Wno-implicit-function-declaration' + \
                ' -fomit-frame-pointer -O2'),
            deps=' '.join([
                'LIBJSON_CFLAGS=" "',
                'LIBJSON_LIBS="-ljson-c"',
                'LIBSNDFILE_CFLAGS=" "',
                'LIBSNDFILE_LIBS="-lsndfile"',
            ]),
            opts=' '.join([
                '--disable-manpages',
                '--disable-neon-opt',
                '--disable-openssl',
                '--disable-orc',
                '--disable-static',
                '--disable-tests',
                '--disable-webrtc-aec',
                '--enable-shared',
                '--without-caps',
            ])))
        execute_make(ctx)
        install_files(ctx, 'config.h', ctx.pkg_inc_dir)
        install_tree(ctx, 'src/pulse', os.path.join(ctx.pkg_inc_dir, 'pulse'),
                     include=['*.h'])
        install_files(ctx, 'src/.libs/libpulse.so', ctx.pkg_lib_dir)
        install_files(ctx, 'src/.libs/libpulse.so.0', ctx.pkg_rpath_dir)
        install_files(ctx, 'src/.libs/libpulse-simple.so', ctx.pkg_lib_dir)
        install_files(ctx, 'src/.libs/libpulse-simple.so.0', ctx.pkg_rpath_dir)
        install_files(ctx, 'src/.libs/libpulsecommon-*.so', ctx.pkg_lib_dir)
        install_files(ctx, 'src/.libs/libpulsecommon-*.so', ctx.pkg_rpath_dir)
    else:
        mkpath('builddir')
        changedir(ctx, 'builddir')
        mkpath('pcdir')
        generate_pc_files(ctx, 'pcdir')
        generate_meson_crossfile(ctx, 'pcdir', 'crossfile.txt')
        execute(ctx, 'meson .. {opts}'.format(
            opts=' '.join([
                '--cross-file=crossfile.txt',
                '-Ddaemon=false',
                '-Ddoxygen=false',
                '-Dgcov=false',
                '-Dman=false',
                '-Dtests=false',
            ])))
        execute(ctx, 'ninja')
        execute(ctx, 'DESTDIR=../instdir ninja install')
        changedir(ctx, '..')
        install_tree(ctx, 'instdir/usr/local/include/pulse',
                     os.path.join(ctx.pkg_inc_dir, 'pulse'),
                     include=['*.h'])
        install_files(ctx, 'builddir/src/pulse/libpulse.so', ctx.pkg_lib_dir)
        install_files(ctx, 'builddir/src/pulse/libpulse.so.0', ctx.pkg_rpath_dir)
        install_files(ctx, 'builddir/src/pulse/libpulse-simple.so', ctx.pkg_lib_dir)
        install_files(ctx, 'builddir/src/pulse/libpulse-simple.so.0', ctx.pkg_rpath_dir)
        install_files(ctx, 'builddir/src/libpulsecommon-*.so', ctx.pkg_lib_dir)
        install_files(ctx, 'builddir/src/libpulsecommon-*.so', ctx.pkg_rpath_dir)
elif ctx.pkg_name == 'ltdl':
    download(
        ctx,
        'https://ftp.gnu.org/gnu/libtool/libtool-{ctx.pkg_ver}.tar.gz',
        'libtool-{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'libtool-{ctx.pkg_ver}.tar.gz',
        'libtool-{ctx.pkg_ver}')
    changedir(ctx, 'src/libtool-{ctx.pkg_ver}')
    execute(ctx, './configure --host={host} {vars} {opts}'.format(
        host=ctx.toolchain,
        vars=format_vars(ctx),
        opts=' '.join([
            '--disable-static',
            '--enable-shared',
        ])))
    execute_make(ctx)
    install_files(ctx, 'libltdl/ltdl.h', ctx.pkg_inc_dir)
    install_tree(ctx, 'libltdl/libltdl', os.path.join(ctx.pkg_inc_dir, 'libltdl'))
    install_files(ctx, 'libltdl/.libs/libltdl.so', ctx.pkg_lib_dir)
    install_files(ctx, 'libltdl/.libs/libltdl.so.*', ctx.pkg_rpath_dir)
elif ctx.pkg_name == 'json-c':
    download(
        ctx,
        'https://github.com/json-c/json-c/archive/json-c-{ctx.pkg_ver}.tar.gz',
        'json-c-{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'json-c-{ctx.pkg_ver}.tar.gz',
        'json-c-json-c-{ctx.pkg_ver}')
    changedir(ctx, 'src/json-c-json-c-{ctx.pkg_ver}')
    execute(ctx, '{configure} --host={host} {vars} {flags} {opts}'.format(
        configure=' '.join(filter(None, [
            # workaround for outdated config.sub
            'ac_cv_host=%s' % ctx.toolchain if ctx.toolchain else '',
            # disable rpl_malloc and rpl_realloc
            'ac_cv_func_malloc_0_nonnull=yes',
            'ac_cv_func_realloc_0_nonnull=yes',
            # configure
            './configure',
        ])),
        host=ctx.toolchain,
        vars=format_vars(ctx),
        flags=format_flags(ctx, cflags='-w -fPIC'),
        opts=' '.join([
            '--enable-static',
            '--disable-shared',
        ])))
    execute_make(ctx, cpu_count=0) # -j is buggy for json-c
    install_tree(ctx, '.', ctx.pkg_inc_dir, include=['*.h'])
    install_files(ctx, '.libs/libjson-c.a', ctx.pkg_lib_dir)
elif ctx.pkg_name == 'gengetopt':
    download(
        ctx,
        'https://ftp.gnu.org/gnu/gengetopt/gengetopt-{ctx.pkg_ver}.tar.gz',
        'gengetopt-{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'gengetopt-{ctx.pkg_ver}.tar.gz',
        'gengetopt-{ctx.pkg_ver}')
    changedir(ctx, 'src/gengetopt-{ctx.pkg_ver}')
    execute(ctx, './configure', clear_env=True)
    execute_make(ctx, cpu_count=0) # -j is buggy for gengetopt
    install_files(ctx, 'src/gengetopt', ctx.pkg_bin_dir)
elif ctx.pkg_name == 'ragel':
    download(
        ctx,
        'https://www.colm.net/files/ragel/ragel-{ctx.pkg_ver}.tar.gz',
        'ragel-{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'ragel-{ctx.pkg_ver}.tar.gz',
        'ragel-{ctx.pkg_ver}')
    changedir(ctx, 'src/ragel-{ctx.pkg_ver}')
    execute(ctx, './configure', clear_env=True)
    execute_make(ctx)
    install_files(ctx, 'ragel/ragel', ctx.pkg_bin_dir)
elif ctx.pkg_name == 'cpputest':
    download(
        ctx,
        'https://github.com/cpputest/cpputest/releases/download/'
            'v{ctx.pkg_ver}/cpputest-{ctx.pkg_ver}.tar.gz',
        'cpputest-{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'cpputest-{ctx.pkg_ver}.tar.gz',
        'cpputest-{ctx.pkg_ver}')
    changedir(ctx, 'src/cpputest-{ctx.pkg_ver}')
    if ctx.prefer_cmake:
        mkpath('build')
        changedir(ctx, 'build')
        execute_cmake(ctx, '..', args=[
            '-DMEMORY_LEAK_DETECTION=OFF',
            '-DTESTS=OFF',
            '-DBUILD_SHARED_LIBS=OFF',
            ])
        execute_cmake_build(ctx)
        shutil.copy('src/CppUTest/libCppUTest.a', '../libCppUTest.a')
        changedir(ctx, '..')
    else:
        execute(ctx, './configure --host={host} {vars} {flags} {opts}'.format(
                host=ctx.toolchain,
                vars=format_vars(ctx),
                # disable warnings, since CppUTest uses -Werror and may fail to
                # build on old GCC versions
                flags=format_flags(ctx, cflags='-w'),
                opts=' '.join([
                    # disable memory leak detection which is too hard to use properly
                    '--disable-memory-leak-detection',
                    # doesn't work on older platforms
                    '--disable-extensions',
                    '--enable-static',
                ])))
        execute_make(ctx)
        shutil.copy('lib/libCppUTest.a', 'libCppUTest.a')
    install_tree(ctx, 'include', ctx.pkg_inc_dir)
    install_files(ctx, 'libCppUTest.a', ctx.pkg_lib_dir)
elif ctx.pkg_name == 'google-benchmark':
    download(
        ctx,
        'https://github.com/google/benchmark/archive/v{ctx.pkg_ver}.tar.gz',
        'benchmark_v{ctx.pkg_ver}.tar.gz')
    unpack(
        ctx,
        'benchmark_v{ctx.pkg_ver}.tar.gz',
        'benchmark-{ctx.pkg_ver}')
    changedir(ctx, 'src/benchmark-{ctx.pkg_ver}')
    if parse_ver(ctx.pkg_ver) < (1, 7, 1) and not which('python') and which('python3'):
        subst_tree(
            ctx, 'tools', ['*.py'],
            from_='#!/usr/bin/env python', to='#!/usr/bin/env python3')
    mkpath('build')
    changedir(ctx, 'build')
    execute_cmake(ctx, '..', args=[
        '-DBENCHMARK_ENABLE_GTEST_TESTS=OFF',
        '-DCMAKE_CXX_FLAGS=-w',
        ])
    execute_cmake_build(ctx)
    changedir(ctx, '..')
    install_tree(ctx, 'include', ctx.pkg_inc_dir, include=['*.h'])
    install_files(ctx, 'build/src/libbenchmark.a', ctx.pkg_lib_dir)
# end of deps
else:
    die("unknown 3rdparty '{}'", ctx.pkg)

# commit successful build
touch(ctx.commit_file)



import SCons.Script
import sys
import re

def _cpu_count():
    try:
        import os
        return len(os.sched_getaffinity(0))
    except:
        pass

    try:
        import multiprocessing
        return multiprocessing.cpu_count()
    except:
        pass

    try:
        import os
        res = int(os.sysconf('SC_NPROCESSORS_ONLN'))
        if res > 0:
            return res
    except:
        pass

    try:
        import os
        res = int(os.environ['NUMBER_OF_PROCESSORS'])
        if res > 0:
            return res
    except:
        pass

    return 1

def init(env):
    for arg in sys.argv:
        if re.match(r'^(-j|--jobs=?)\d*$', arg):
            return
    SCons.Script.SetOption('num_jobs', _cpu_count())



from __future__ import print_function

import argparse
import glob
import os.path
import shutil
import subprocess
import sys
import tempfile

try:
    from shlex import quote
except:
    from pipes import quote

VERBOSE = False

def path_dirname(path):
    return os.path.dirname(path)

def path_basename(path):
    return os.path.splitext(os.path.basename(path))[0]

def is_gnu_tool(tool):
    try:
        out = subprocess.check_output([tool, '-V'], stderr=subprocess.STDOUT)
        try:
            out = out.decode()
        except:
            pass
        try:
            out = str(out)
        except:
            pass
        return 'GNU' in out
    except:
        return False

def execute_command(cmd):
    if VERBOSE:
        print(cmd, file=sys.stderr)
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    out = proc.stdout.read().decode()
    code = proc.wait()
    if code != 0:
        print(cmd, file=sys.stderr)
        print(out, file=sys.stderr)
        print("command exited with code {}".format(code))
        exit(1)

parser = argparse.ArgumentParser(description='compose multiple static libs into one')

parser.add_argument('--out', dest='output', type=str, required=True,
                    help='output library')

parser.add_argument('--in', dest='inputs', type=str, nargs='+', required=True,
                    help='input libraries')

parser.add_argument('--arch', dest='arch', type=str, nargs='*',
                    help='binary achitecture(s)')

parser.add_argument('--tools', dest='tools', type=str, nargs='*',
                    help='paths to tools')

parser.add_argument('--verbose', dest='verbose', action='store_true',
                    help='enable verbose output')

args = parser.parse_args()

VERBOSE = args.verbose

dst_lib = os.path.abspath(args.output)
src_libs = [os.path.abspath(arg) for arg in args.inputs]

archs = args.arch or [None]
is_fat = len(archs) > 1

tools = dict()
for e in args.tools:
    k, v = e.split('=', 1)
    tools[k] = v

for exe in ['ar', 'objcopy', 'lipo']:
    if not tools.get(exe.upper(), None):
        tools[exe.upper()] = exe

have_gnu_objcopy = is_gnu_tool(tools['OBJCOPY'])

try:
    temp_dir = tempfile.mkdtemp()
    os.chdir(temp_dir)

    # remove output lib(s)
    for path in glob.glob(dst_lib+'*'):
        os.remove(path)

    for arch in archs:
        dst_arch_lib = '{}/{}.{}.a'.format(
            path_dirname(dst_lib),
            path_basename(dst_lib),
            arch)

        for path in glob.glob(dst_arch_lib+'*'):
            os.remove(path)

    for src_lib in src_libs:
        for arch in archs:
            basename = path_basename(src_lib)

            # unpack object files for given arch from input lib
            if is_fat:
                arch_lib = '{}.{}.a'.format(basename, arch)

                execute_command(
                    '{lipo} -extract_family {arch} -output {arch_lib} {src_lib}'.format(
                        lipo=quote(tools['LIPO']),
                        arch=quote(arch),
                        arch_lib=quote(arch_lib),
                        src_lib=quote(src_lib)))

                execute_command(
                    '{ar} x {arch_lib}'.format(
                        ar=quote(tools['AR']),
                        arch_lib=quote(arch_lib)))

                os.remove(arch_lib)
            else:
                execute_command(
                    '{ar} x {src_lib}'.format(
                        ar=quote(tools['AR']),
                        src_lib=quote(src_lib)))

            for obj in list(glob.glob('*.o')):
                new_obj = '{}-{}'.format(basename, obj)

                # transform object file
                if have_gnu_objcopy:
                    execute_command(
                        '{objcopy} --localize-hidden --strip-unneeded {obj} {new_obj}'.format(
                            objcopy=quote(tools['OBJCOPY']),
                            obj=quote(obj),
                            new_obj=quote(new_obj)))
                    os.remove(obj)
                else:
                    os.rename(obj, new_obj)

                # add object file to resulting static lib for given arch
                if is_fat:
                    dst_arch_lib = '{}/{}.{}.a'.format(
                        path_dirname(dst_lib),
                        path_basename(dst_lib),
                        arch)

                    # quick mode (disable checks, disable symbol index)
                    execute_command(
                        '{ar} qS {dst_arch_lib} {new_obj}'.format(
                            ar=quote(tools['AR']),
                            dst_arch_lib=quote(dst_arch_lib),
                            new_obj=quote(new_obj)))
                else:
                    execute_command(
                        '{ar} r {dst_lib} {new_obj}'.format(
                            ar=quote(tools['AR']),
                            dst_lib=quote(dst_lib),
                            new_obj=quote(new_obj)))

                os.remove(new_obj)

    # compose per-arch static libs into one
    if is_fat:
        dst_arch_libs = [
            '{}/{}.{}.a'.format(
                    path_dirname(dst_lib),
                    path_basename(dst_lib),
                    arch)
            for arch in archs]

        # build symbol index
        for dst_arch_lib in dst_arch_libs:
            execute_command(
                '{ar} s {dst_arch_lib}'.format(
                    ar=quote(tools['AR']),
                    dst_arch_lib=quote(dst_arch_lib)))

        execute_command(
              '{lipo} -create -output {dst_lib} {dst_arch_libs}'.format(
                  lipo=quote(tools['LIPO']),
                  dst_lib=quote(dst_lib),
                  dst_arch_libs=' '.join(map(quote, dst_arch_libs))))

except:
    if os.path.isfile(dst_lib):
        os.remove(dst_lib)
    raise

finally:
    if os.path.isdir(temp_dir):
        shutil.rmtree(temp_dir)



# -*- coding: utf-8 -*-

import datetime
import os
import re
import subprocess
import sys

def get_version():
    proc = subprocess.Popen(
        [sys.executable,
         os.path.dirname(os.path.realpath(__file__)) +
            '/../../scripts/scons_helpers/parse-version.py'],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT)
    return tuple(proc.stdout.read().decode().strip().split('.'))

# -- General configuration ------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinxemoji.sphinxemoji',
    'breathe',
]

templates_path = []

source_suffix = ['.rst']
exclude_patterns = []

master_doc = 'index'

project = u'Roc Toolkit'
copyright = u'{}, Roc Streaming authors'.format(datetime.datetime.now().year)
author = u'Roc Streaming authors'

version_tuple = get_version()

version = 'Roc Toolkit {}'.format('.'.join(version_tuple[:2]))
release = '.'.join(version_tuple)

today_fmt = '%Y'

pygments_style = 'sphinx'

todo_include_todos = False

# -- Options for Breathe ----------------------------------------------

breathe_projects = { 'roc': '../../build/docs/public_api/xml' }

breathe_default_project = 'roc'
breathe_domain_by_extension = {'h': 'c'}

# -- Options for HTML output ----------------------------------------------

html_title = '{} {}'.format(project, release)

html_theme = 'nature'

html_logo = '../images/logo80.png'

html_sidebars = {
   '**': ['globaltoc.html', 'searchbox.html'],
}

html_static_path = ['_static']

html_css_files = [
    'roc.css',
]

html_js_files = [
    'analytics.js',
]

# -- Options for manual page output ---------------------------------------

man_pages = [
    ('manuals/roc_send', 'roc-send', u'send real-time audio', [], 1),
    ('manuals/roc_recv', 'roc-recv', u'receive real-time audio', [], 1),
    ('manuals/roc_copy', 'roc-copy', u'copy local audio', [], 1),
]



#! /usr/bin/env python3

import fileinput
import re

def read_lines():
    for line in fileinput.input():
        yield line.strip()

def filter_options(inp):
    for line in inp:
        if re.match(r'^\s*\-', line):
            yield line

def concat_options(inp):
    full_line = None
    for line in inp:
        if line.startswith('-') or line == '':
            if full_line:
                yield full_line
            full_line = line
        else:
            if full_line:
                full_line += ' ' + line
    yield full_line

def format_options(inp):
    maxlen = 0

    col1 = []
    col2 = []

    for line in inp:
        m = re.match(r'^((\-\w,\s)?\-\-\S+)\s+(.+)$', line)
        if m:
            maxlen = max(maxlen, len(m.group(1)))
            col1.append(m.group(1))
            col2.append(m.group(3))

    for c1, c2 in zip(col1, col2):
        print(('{:<' + str(maxlen + 2) + '}{}').format(c1, c2))

format_options(
    filter_options(
        concat_options(
            read_lines())))
